'''experiment.py - writing reproducible scripts
=========================================================

The :mod:`experiment` modules contains utility functions for argument
parsing, logging and record keeping within scripts.

This module is imported by most CGAT scripts. It provides convenient
and consistent methods for

   * `Record keeping`_
   * `Argument parsing`_
   * `Input/Output redirection`_
   * `Logging`_
   * `Running external commands`_
   * `Benchmarking`_

See :doc:`../scripts/cgat_script_template` on how to use this module.

This module can handle both optparse and argparse. The default is to
return an argparse object.

For default argparse: The basic usage of this module within a script is::

    """script_name.py - my script

    Mode Documentation
    """
    import sys
    import optparse
    import CGAT.experiment as E

    def main(argv=None):
        """script main.

        parses command line options in sys.argv, unless *argv* is given.
        """

        if not argv: argv = sys.argv

        # setup command line parser
        parser = E.OptionParser(description=__doc__)

        parser.add_arguments("-t", "--test", dest="test", type="string",
                          help="supply help")

        # add common options (-h/--help, ...) and parse
        # command line
        args = E.Start(parser)

        # do something
        # ...
        E.info("an information message")
        E.warn("a warning message)

        ## write footer and output benchmark information.
        E.Stop()

    if __name__ == "__main__":
        sys.exit(main(sys.argv))


To use optparse: The basic usage of this module within a script is::

    def main(argv=None):
        """script main.

        parses command line options in sys.argv, unless *argv* is given.
        """

        if not argv: argv = sys.argv

        # setup command line parser
        parser = E.OptionParser(version="%prog version: $Id$",
                                usage=globals()["__doc__"], optparse=True)

        parser.add_option("-t", "--test", dest="test", type="string",
                          help="supply help")

        # add common options (-h/--help, ...) and parse
        # command line
        (options, args) = E.Start(parser, optparse=True)

        # do something
        # ...
        E.info("an information message")
        E.warn("a warning message)

        ## write footer and output benchmark information.
        E.Stop()

    if __name__ == "__main__":
        sys.exit(main(sys.argv))


Record keeping
--------------

The central functions in this module are the :py:func:`Start` and
:py:func:`Stop` methods which are called before or after any work is
done within a script.

The :py:func:`Start` is called with an E.OptionParser object.
:py:func:`Start` will add additional command line arguments, such as
``--help`` for command line help or ``--verbose`` to control the
:term:`loglevel`.  It can also add optional arguments for scripts
needing database access, writing to multiple output files, etc.

:py:func:`Start` will write record keeping information to a
logfile. Typically, logging information is output on stdout, prefixed
by a `#`, but it can be re-directed to a separate file. Below is a
typical output::

    # output generated by /ifs/devel/andreas/cgat/beds2beds.py --force-output --exclusive-overlap --method=unmerged-combinations --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz Rela-030m-R1.bed.gz  # nopep8
    # job started at Thu Mar 29 13:06:33 2012 on cgat150.anat.ox.ac.uk -- e1c16e80-03a1-4023-9417-f3e44e33bdcd
    # pid: 16649, system: Linux 2.6.32-220.7.1.el6.x86_64 #1 SMP Fri Feb 10 15:22:22 EST 2012 x86_64
    # exclusive                               : True
    # filename_update                         : None
    # ignore_strand                           : False
    # loglevel                                : 1
    # method                                  : unmerged-combinations
    # output_filename_pattern                 : 030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
    # output_force                            : True
    # pattern_id                              : (.*).bed.gz
    # stderr                                  : <open file \'<stderr>\', mode \'w\' at 0x2ba70e0c2270>
    # stdin                                   : <open file \'<stdin>\', mode \'r\' at 0x2ba70e0c2150>
    # stdlog                                  : <open file \'030m.intersection.tsv.log\', mode \'a\' at 0x1f1a810>
    # stdout                                  : <open file \'<stdout>\', mode \'w\' at 0x2ba70e0c21e0>
    # timeit_file                             : None
    # timeit_header                           : None
    # timeit_name                             : all
    # tracks                                  : None

The header contains information about:

    * the script name (``beds2beds.py``)

    * the command line options (``--force-output --exclusive-overlap
      --method=unmerged-combinations
      --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
      --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz
      Rela-030m-R1.bed.gz``)

    * the time when the job was started (``Thu Mar 29 13:06:33 2012``)
    * the location it was executed (``cgat150.anat.ox.ac.uk``)
    * a unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)
    * the pid of the job (``16649``)

    * the system specification (``Linux 2.6.32-220.7.1.el6.x86_64 #1
      SMP Fri Feb 10 15:22:22 EST 2012 x86_64``)

It is followed by a list of all options that have been set in the script.

Once completed, a script will call the :py:func:`Stop` function to
signify the end of the experiment.

:py:func:`Stop` will output to the log file that the script has
concluded successfully. Below is typical output::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 \
      -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

The footer contains information about:

   * the job has finished (``job finished``)
   * the time it took to execute (``11 seconds``)
   * when it completed (``Thu Mar 29 13:06:44 2012``)
   * some benchmarking information (``11.36  0.45  0.00  0.01``)
     which is ``user time``, ``system time``,
     ``child user time``, ``child system time``.
   * the unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)

The unique job id can be used to easily retrieve matching information
from a concatenation of log files.

Argument parsing
----------------

The module provides :class:`OptionParser` to facilitate option
parsing.  :class:`OptionParser` is derived from the
:py:class:`optparse.OptionParser` class, but has improvements to
provide better formatted output on the command line. It also allows to
provide a comma-separated list to options that accept multiple
arguments. Thus, ``--method=sort --method=crop`` and
``--method=sort,crop`` are equivalent.

Additionally, there are set of commonly used option groups that are
used in many scripts. The :func:`Start` method has options to automatically
add these. For example::

   (options, args) = E.Start(parser, add_output_options=True)

will add the option ``--output-filename-pattern``. Similarly::

   (options, args) = E.Start(parser, add_database_options=True)

will add multiple options for scripts accessing databases, such as
``--database-host`` and ``--database-username``.

Input/Output redirection
------------------------

:func:`Start` adds the options ``--stdin``, ``--stderr` and
``--stdout`` which allow using files as input/output streams.

To make this work, scripts should not read from sys.stdin or write to
sys.stdout directly, but instead use ``options.stdin`` and
``options.stdout``. For example to simply read all lines from stdin
and write to stdout, use::

   (options, args) = E.Start(parser)

   input_data = options.stdin.readlines()
   options.stdout.write("".join(input_data))

The script can then be used in many different contexts::

   cat in.data | python script.py > out.data
   python script.py --stdin=in.data > out.data
   python script.py --stdin=in.data --stdout=out.data

The method handles gzip compressed files transparently. The following
are equivalent::

   zcat in.data.gz | python script.py | gzip > out.data.gz
   python script.py --stdin=in.data.gz --stdout=out.data.gz

For scripts producing multiple output files, use the argument
``add_output_options=True`` to :func:`Start`. This provides the option
``--output-filename-pattern`` on the command line. The user can then
supply a pattern for output files. Any ``%s`` appearing in the pattern
will be substituted by a ``section``. Inside the script, When opening
an output file, use the method :func:`open_output_file` to provide a
file object::

   output_histogram = E.open_output_file(section="histogram")
   output_stats = E.open_output_file(section="stats")

If the user calls the script with::

   python script.py --output-filename-pattern=sample1_%s.tsv.gz

the script will create the files ``sample1_histogram.tsv.gz`` and
``sample1_stats.tsv.gz``.

This method will also add the option ``--force-output`` to permit
overwriting existing files.

Logging
-------

:py:mod:`experiment` provides the well known logging methods from
the :py:mod:`logging` module such as :py:func:`info`,
:py:func:`warn`, etc. These are provided so that no additional import
of the :py:mod:`logging` module is required, but either functions
can be used.

Running external commands
-------------------------

The :func:`run` method is a shortcut :py:func:`subprocess.call` and
similar methods with some additional sanity checking.

Benchmarking
------------

The :func:`Start` method records basic benchmarking information when a
script starts and :func:`Stop` outputs it as part of its final log
message::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 \
      -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

See `Record keeping`_ for an explanations of the fields.

To facilitate collecting benchmark information from running multiple
scripts, these data can be tagged and saved in a separate file. See the
command line options ``--timeit``, ``--timeit-name``, ``--timeit-header``
in :func:`Start`.

The module contains some decorator functions for benchmarking
(:func:`benchmark`) and caching function (:func:`cached_function`) or
class method (:func:`cached_method`) calls.

API
---

'''

import re
import sys
import time
import inspect
import copy
import os
import collections
import subprocess
import functools
import gzip
import warnings
import pipes
import optparse
import argparse
import textwrap
import random
import uuid
import yaml
# import convenience functions from logging
import logging
import logging.config
from logging import warning, info, log, debug, error, critical, exception
from logging import warning as warn


class DefaultOptions:
    stdlog = sys.stdout
    stdout = sys.stdout
    stderr = sys.stderr
    stdin = sys.stdin
    loglevel = 2
    timeit_file = None


global_starting_time = time.time()
global_options = DefaultOptions()
global_args = None
global_id = uuid.uuid4()
global_benchmark = collections.defaultdict(int)

##########################################################################
# The code for BetterFormatter has been taken from
# http://code.google.com/p/yjl/source/browse/Python/snippet/BetterFormatter.py
__copyright__ = """
Copyright (c) 2001-2006 Gregory P. Ward.  All rights reserved.
Copyright (c) 2002-2006 Python Software Foundation.  All rights reserved.
Copyright (c) 2011 Yu-Jie Lin.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  * Neither the name of the author nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class BetterFormatter(optparse.IndentedHelpFormatter):
    """A formatter for :class:`OptionParser` outputting indented
    help text.
    """

    def __init__(self, *args, **kwargs):

        optparse.IndentedHelpFormatter.__init__(self, *args, **kwargs)
        self.wrapper = textwrap.TextWrapper(width=self.width)

    def _formatter(self, text):

        return '\n'.join(['\n'.join(p) for p in
                          map(self.wrapper.wrap,
                              self.parser.expand_prog_name(text).split('\n'))])

    def format_description(self, description):

        if description:
            return self._formatter(description) + '\n'
        else:
            return ''

    def format_epilog(self, epilog):

        if epilog:
            return '\n' + self._formatter(epilog) + '\n'
        else:
            return ''

    def format_usage(self, usage):

        return self._formatter(optparse._("Usage: %s\n") % usage)

    def format_option(self, option):
        # Ripped and modified from Python 2.6's optparse's HelpFormatter
        result = []
        opts = self.option_strings[option]
        opt_width = self.help_position - self.current_indent - 2
        if len(opts) > opt_width:
            opts = "%*s%s\n" % (self.current_indent, "", opts)
            indent_first = self.help_position
        else:                       # start help on same line as opts
            opts = "%*s%-*s  " % (self.current_indent, "", opt_width, opts)
            indent_first = 0
        result.append(opts)
        if option.help:
            help_text = self.expand_default(option)
            # Added expand program name
            help_text = self.parser.expand_prog_name(help_text)
            # Modified the generation of help_line
            help_lines = []
            wrapper = textwrap.TextWrapper(width=self.help_width)
            for p in map(wrapper.wrap, help_text.split('\n')):
                if p:
                    help_lines.extend(p)
                else:
                    help_lines.append('')
            # End of modification
            result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            result.extend(["%*s%s\n" % (self.help_position, "", line)
                           for line in help_lines[1:]])
        elif opts[-1] != "\n":
            result.append("\n")
        return "".join(result)


# End of BetterFormatter()
#################################################################
#################################################################
#################################################################

class AppendCommaOption(optparse.Option):

    '''Option with additional parsing capabilities.

    * "," in arguments to options that have the action 'append'
      are treated as a list of options. This is what galaxy does,
      but generally convenient.

    * Option values of "None" and "" are treated as default values.
    '''
#    def check_value( self, opt, value ):
# do not check type for ',' separated lists
#        if "," in value:
#            return value
#        else:
#            return optparse.Option.check_value( self, opt, value )
#
#    def take_action(self, action, dest, opt, value, values, parser):
#        if action == "append" and "," in value:
#            lvalue = value.split(",")
#            values.ensure_value(dest, []).extend(lvalue)
#        else:
#            optparse.Option.take_action(
#                self, action, dest, opt, value, values, parser)
#

    def convert_value(self, opt, value):
        if value is not None:
            if self.nargs == 1:
                if self.action == "append":
                    if "," in value:
                        return [self.check_value(opt, v) for v in
                                value.split(",") if v != ""]
                    else:
                        if value != "":
                            return self.check_value(opt, value)
                        else:
                            return value
                else:
                    return self.check_value(opt, value)
            else:
                return tuple([self.check_value(opt, v) for v in value])

    # why is it necessary to pass action and dest to this function when
    # they could be accessed as self.action and self.dest?
    def take_action(self, action, dest, opt, value, values, parser):

        if action == "append" and type(value) == list:
            values.ensure_value(dest, []).extend(value)
        else:
            optparse.Option.take_action(
                self, action, dest, opt, value, values, parser)


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


class ArgumentParser(argparse.ArgumentParser):
    ''''CGAT derivative of ArgumentParser. OptionParser is still
    implimented for backwards compatibility
    '''

    def __init__(self, *args, **kwargs):

        # if "--short" is a command line option
        # remove usage from kwargs
        if "--no-usage" in sys.argv:
            kwargs["usage"] = None

        argparse.ArgumentParser.__init__(self, *args,
                                         formatter_class=CustomFormatter,
                                         **kwargs)

        if "--no-usage" in sys.argv:
            self.add_argument("--no-usage", dest="help_no_usage",
                              action="store_true",
                              help="output help without usage information")


class OptionParser(optparse.OptionParser):
    '''CGAT derivative of ArgumentParser. OptionParser is still
    implemented for backwards compatibility

    '''

    def __init__(self, *args, **kwargs):

        # if "--short" is a command line option
        # remove usage from kwargs
        if "--no-usage" in sys.argv:
            kwargs["usage"] = None

        optparse.OptionParser.__init__(self, *args,
                                       option_class=AppendCommaOption,
                                       formatter=BetterFormatter(),
                                       **kwargs)

        # set new option parser
        # parser.formatter = BetterFormatter()
        # parser.formatter.set_parser(parser)
        if "--no-usage" in sys.argv:
            self.add_option("--no-usage", dest="help_no_usage",
                            action="store_true",
                            help="output help without usage information")


class OptionGroup(optparse.OptionGroup):
    pass


def callbackShortHelp(option, opt, value, parser):
    '''output short help (only command line options).'''
    # clear usage and description
    parser.set_description(None)
    parser.set_usage(None)
    # output help
    parser.print_help()
    # exit
    parser.exit()


def open_file(filename, mode="r", create_dir=False, encoding="utf-8"):
    '''open file in *filename* with mode *mode*.

    If *create* is set, the directory containing filename
    will be created if it does not exist.

    gzip - compressed files are recognized by the
    suffix ``.gz`` and opened transparently.

    Note that there are differences in the file
    like objects returned, for example in the
    ability to seek.

    returns a file or file-like object.
    '''

    _, ext = os.path.splitext(filename)

    if create_dir:
        dirname = os.path.abspath(os.path.dirname(filename))
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

    if ext.lower() in (".gz", ".z"):
        # in gzip, r defaults to "rt", so make it compatible with open
        if mode == "r":
            mode = "rt"
        elif mode == "w":
            mode = "wt"
        return gzip.open(filename, mode, encoding=encoding)
    else:
        return open(filename, mode, encoding=encoding)


def get_header():
    """return a header string with command line options and timestamp

    """
    system, host, release, version, machine = os.uname()

    return "output generated by %s\njob started at %s on %s -- %s\npid: %i, system: %s %s %s %s" %\
           (" ".join(sys.argv),
            time.asctime(time.localtime(time.time())),
            host,
            global_id,
            os.getpid(),
            system, release, version, machine)


def get_params(options=None):
    """return a string containing script parameters.

    Parameters are all variables that start with ``param_``.
    """
    result = []
    if options:
        members = options.__dict__
        for k, v in sorted(members.items()):
            result.append("%-40s: %s" % (k, str(v)))
    else:
        vars = inspect.currentframe().f_back.f_locals
        for var in [x for x in list(vars.keys()) if re.match("param_", x)]:
            result.append("%-40s: %s" %
                          (var, str(vars[var])))

    if result:
        return "\n".join(result)
    else:
        return "# no parameters."


def get_footer():
    """return a header string with command line options and
    timestamp.
    """
    return "job finished in %i seconds at %s -- %s -- %s" %\
           (time.time() - global_starting_time,
            time.asctime(time.localtime(time.time())),
            " ".join(["%5.2f" % x for x in os.times()[:4]]),
            global_id)


class MultiLineFormatter(logging.Formatter):
    '''logfile formatter: add identation for multi-line entries.'''

    def format(self, record):

        s = logging.Formatter.format(self, record)
        if s.startswith("#"):
            prefix = "#"
        else:
            prefix = ""
        if record.message:
            header, footer = s.split(record.message)
            s = s.replace("\n", " \\\n%s" % prefix + " " * (len(header) - 1))
        return s


def get_args():
    """return global options depending on parser chosen.
    """
    if isinstance(global_args, list):
        return global_options
    return global_args


def _setup_seed(options):
    if options.random_seed is not None:
        random.seed(options.random_seed)


def _setup_pipes(options, add_pipe_options):
    if add_pipe_options:
        if options.stdout != sys.stdout:
            options.stdout = open_file(options.stdout, "w")
        if options.stderr != sys.stderr:
            if options.stderr == "stderr":
                options.stderr = options.stderr
            else:
                options.stderr = open_file(
                    options.stderr, "w")
        if options.stdlog != sys.stdout:
            options.stdlog = open_file(options.stdlog, "a")
        if options.stdin != sys.stdin:
            options.stdin = open_file(options.stdin, "r")
    else:
        options.stderr = sys.stderr
        options.stdout = sys.stdout
        options.stdlog = sys.stdout
        options.stdin = sys.stdin


def _setup_logging(options, logger_callback):
    # reset log_config_filename if logging.yml does not exist
    if options.log_config_filename == "logging.yml" and not os.path.exists(
            options.log_config_filename):
        options.log_config_filename = None

    if options.log_config_filename:
        if os.path.exists(options.log_config_filename):
            # configure logging from filename
            with open(options.log_config_filename) as inf:
                dict_yaml = yaml.safe_load(inf)
            logging.config.dictConfig(dict_yaml)
        else:
            raise OSError("file {} with logging configuration does not exist".format(
                options.log_config_filename))
    else:
        # configure logging from command line options
        # map from 0-10 to logging scale
        # 0: quiet
        # 1: little verbositiy
        # >1: increased verbosity
        if options.loglevel == 0:
            lvl = logging.ERROR
        elif options.loglevel == 1:
            lvl = logging.INFO
        else:
            lvl = logging.DEBUG

        if options.stdout == options.stdlog:
            logformat = '# %(asctime)s %(levelname)s %(message)s'
        else:
            logformat = '%(asctime)s %(levelname)s %(message)s'

        logging.basicConfig(
            level=lvl,
            format=logformat,
            stream=options.stdlog)

        # set up multi-line logging
        # Note that .handlers is not part of the API, might change
        # Solution is to configure handlers explicitely.
        for handler in logging.getLogger().handlers:
            handler.setFormatter(MultiLineFormatter(logformat))

    if logger_callback:
        logger = logger_callback(options)
    else:
        logger = logging.getLogger("cgatcore")
    return logger


def _setup_tracing(options):
    if options.tracing == "function":
        sys.settrace(trace_calls)


def start(parser=None,
          argv=None,
          quiet=False,
          no_parsing=False,
          add_csv_options=False,
          add_database_options=False,
          add_pipe_options=True,
          add_cluster_options=False,
          add_output_options=False,
          logger_callback=None,
          return_parser=False,
          unknowns=False):
    """set up an experiment.

    The :py:func:`Start` method will set up a file logger and add some
    default and some optional options to the command line parser.  It
    will then parse the command line and set up input/output
    redirection and start a timer for benchmarking purposes.

    The default options added by this method are:

    ``-v/--verbose``
        the :term:`loglevel`

    ``timeit``
        turn on benchmarking information and save to file

    ``timeit-name``
         name to use for timing information,

    ``timeit-header``
         output header for timing information.

    ``seed``
         the random seed. If given, the python random
         number generator will be initialized with this
         seed.

    Optional options added are:

    add_csv_options

       ``dialect``
            csv_dialect. the default is ``excel-tab``, defaulting to
            :term:`tsv` formatted files.

    add_database_options
       ``-C/--connection``
           psql connection string
       ``-U/--user``
           psql user name

    add_cluster_options
       ``--use-cluster``
           use cluster
       ``--cluster-priority``
           cluster priority to request
       ``--cluster-queue``
           cluster queue to use
       ``--cluster-num-jobs``
           number of jobs to submit to the cluster at the same time
       ``--cluster-options``
           additional options to the cluster for each job.

    add_output_options
       ``-P/--output-filename-pattern``
            Pattern to use for output filenames.

    Arguments
    ---------

    param parser : :py:class:`E.OptionParser`
       instance with command line options.

    argv : list
        command line options to parse. Defaults to
        :py:data:`sys.argv`

    quiet : bool
        set :term:`loglevel` to 0 - no logging

    no_parsing : bool
        do not parse command line options

    return_parser : bool
        return the parser object, no parsing. Useful for inspecting
        the command line options of a script without running it.

    add_csv_options : bool
        add common options for parsing :term:`tsv` separated files.

    add_database_options : bool
        add common options for connecting to various databases.

    add_pipe_options : bool
        add common options for redirecting input/output

    add_cluster_options : bool
        add common options for scripts submitting jobs to the cluster

    add_output_options : bool
        add commond options for working with multiple output files

    logger_callback : object
        callback function to further configure logging system. The
        callback should accept the options as first parameter and
        return a logger.

    unknowns : bool
        if a set of unknown args are to be returned

    optparse : bool
        specify if parser type is either optparse or argparse

    Returns
    -------
    tuple
       (:py:class:`E.OptionParser` object, list of positional
       arguments)

    """

    if argv is None:
        argv = sys.argv

    global global_args, global_starting_time

    global_starting_time = time.time()

    # Argparse options
    if "OptionParser" in str(parser.__class__):
        if not parser:
            parser = OptionParser(version="%prog version: $Id$")

        global global_options

        # save default values given by user
        user_defaults = copy.copy(parser.defaults)

        group = OptionGroup(parser, "Script timing options")

        group.add_option("--timeit", dest='timeit_file', type="string",
                         help="store timeing information in file [%default].")
        group.add_option("--timeit-name", dest='timeit_name', type="string",
                         help="name in timing file for this class of jobs "
                         "[%default].")
        group.add_option("--timeit-header", dest='timeit_header',
                         action="store_true",
                         help="add header for timing information [%default].")
        parser.add_option_group(group)

        group = OptionGroup(parser, "Common options")

        group.add_option("--random-seed", dest='random_seed', type="int",
                         help="random seed to initialize number generator "
                         "with [%default].")

        group.add_option("-v", "--verbose", dest="loglevel", type="int",
                         help="loglevel [%default]. The higher, the more output.")

        group.add_option("--log-config-filename",
                         dest="log_config_filename",
                         type="string",
                         default="logging.yml",
                         help="Configuration file for logger [%default].")

        group.add_option("--tracing", dest="tracing", type="choice",
                         choices=("function",),
                         default=None,
                         help="enable function tracing [%default].")

        group.add_option("-?", dest="short_help", action="callback",
                         callback=callbackShortHelp,
                         help="output short help (command line options only.")

        parser.add_option_group(group)

        if quiet:
            parser.set_defaults(loglevel=0)
        else:
            parser.set_defaults(loglevel=1)

        parser.set_defaults(
            timeit_file=None,
            timeit_name='all',
            timeit_header=None,
            random_seed=None,
            tracing=None,
        )

        if add_csv_options:
            parser.add_option("--csv-dialect", dest="csv_dialect", type="string",
                              help="csv dialect to use [%default].")

            parser.set_defaults(
                csv_dialect="excel-tab",
                csv_lineterminator="\n",
            )

        if add_cluster_options:
            group = OptionGroup(parser, "cluster options")
            group.add_option("--no-cluster", "--local", dest="without_cluster",
                             action="store_true",
                             help="do no use cluster - run locally [%default].")
            group.add_option("--cluster-priority", dest="cluster_priority",
                             type="int",
                             help="set job priority on cluster [%default].")
            group.add_option("--cluster-queue", dest="cluster_queue",
                             type="string",
                             help="set cluster queue [%default].")
            group.add_option("--cluster-num-jobs", dest="cluster_num_jobs",
                             type="int",
                             help="number of jobs to submit to the queue execute "
                             "in parallel [%default].")
            group.add_option("--cluster-parallel-environment",
                             dest="cluster_parallel_environment",
                             type="string",
                             help="name of the parallel environment to use "
                             "[%default].")
            group.add_option("--cluster-options", dest="cluster_options",
                             type="string",
                             help="additional options for cluster jobs, passed "
                             "on to queuing system [%default].")
            group.add_option("--cluster-queue-manager",
                             dest="cluster_queue_manager",
                             type="choice",
                             choices=("sge", "slurm", "torque", "pbspro"),
                             help="cluster queuing system "
                             "[%default].")
            group.add_option("--cluster-memory-resource",
                             dest="cluster_memory_resource",
                             type="string",
                             help="resource name to allocate memory with "
                             "[%default].")
            group.add_option("--cluster-memory-default",
                             dest="cluster_memory_default",
                             type="string",
                             help="default amount of memory to allocate "
                             "[%default].")

            parser.set_defaults(without_cluster=False,
                                cluster_queue=None,
                                cluster_priority=None,
                                cluster_num_jobs=None,
                                cluster_parallel_environment=None,
                                cluster_options=None,
                                cluster_memory_resource=None,
                                cluster_memory_default="unlimited",
                                cluster_queue_manager="sge")
            parser.add_option_group(group)

        if add_output_options or add_pipe_options:
            group = OptionGroup(parser, "Input/output options")

            if add_output_options:
                group.add_option(
                    "-P", "--output-filename-pattern",
                    dest="output_filename_pattern", type="string",
                    help="OUTPUT filename pattern for various methods "
                    "[%default].")

                group.add_option("-F", "--force-output", dest="output_force",
                                 action="store_true",
                                 help="force over-writing of existing files.")

                parser.set_defaults(output_filename_pattern="%s",
                                    output_force=False)

            if add_pipe_options:

                group.add_option("-I", "--stdin", dest="stdin", type="string",
                                 help="file to read stdin from [default = stdin].",
                                 metavar="FILE")
                group.add_option("-L", "--log", dest="stdlog", type="string",
                                 help="file with logging information "
                                 "[default = stdout].",
                                 metavar="FILE")
                group.add_option("-E", "--error", dest="stderr", type="string",
                                 help="file with error information "
                                 "[default = stderr].",
                                 metavar="FILE")
                group.add_option("-S", "--stdout", dest="stdout", type="string",
                                 help="file where output is to go "
                                 "[default = stdout].",
                                 metavar="FILE")

                parser.set_defaults(stderr=sys.stderr)
                parser.set_defaults(stdout=sys.stdout)
                parser.set_defaults(stdlog=sys.stdout)
                parser.set_defaults(stdin=sys.stdin)

            parser.add_option_group(group)

        if add_database_options:
            group = OptionGroup(parser, "database connection options")
            group.add_option(
                "--database-url", dest="database_url", type="string",
                help="database connection url, for example sqlite:///./csvdb [%default].")

            group.add_option(
                "--database-schema", dest="database_schema", type="string",
                help="database schema [%default]")

            parser.set_defaults(database_url="sqlite:///./csvdb")
            parser.set_defaults(database_schema=None)
            parser.add_option_group(group)

        # restore user defaults
        parser.defaults.update(user_defaults)

        if return_parser:
            return parser

        if not no_parsing:
            (global_options, global_args) = parser.parse_args(argv[1:])

        opt1, opt2 = global_options, global_args
        retval = (opt1, opt2)

    # Argparse options
    else:
        group = parser.add_argument_group("Script timing options")

        group.add_argument("--timeit", dest='timeit_file', type=str,
                           help="store timeing information in file.")
        group.add_argument("--timeit-name", dest='timeit_name', type=str,
                           help="name in timing file for this class of jobs ")
        group.add_argument("--timeit-header", dest='timeit_header',
                           action="store_true",
                           help="add header for timing information.")

        group = parser.add_argument_group("Common options")

        group.add_argument("--random-seed", dest='random_seed', type=int,
                           help="random seed to initialize number generator "
                           "with")

        group.add_argument("-v", "--verbose", dest="loglevel", type=int,
                           help="loglevel. The higher, the more output.")

        group.add_argument("--log-config-filename",
                           dest="log_config_filename",
                           type=str,
                           default="logging.yml",
                           help="Configuration file for logger.")

        group.add_argument("--tracing", dest="tracing", type=str,
                           choices=["function"],
                           default=None,
                           help="enable function tracing.")

        group.add_argument("-?", type=callbackShortHelp,
                           help="output short help (command line options only.")

        if quiet:
            parser.set_defaults(loglevel=0)
        else:
            parser.set_defaults(loglevel=1)

        parser.set_defaults(
            timeit_file=None,
            timeit_name='all',
            timeit_header=None,
            random_seed=None,
            tracing=None,
        )

        if add_csv_options:
            parser.add_argument("--csv-dialect", dest="csv_dialect", type=str,
                                help="csv dialect to use.")

            parser.set_defaults(
                csv_dialect="excel-tab",
                csv_lineterminator="\n",
            )

        if add_cluster_options:
            group = parser.add_argument_group("cluster options")
            group.add_argument("--no-cluster", "--local", dest="without_cluster",
                               action="store_true",
                               help="do no use cluster - run locally.")
            group.add_argument("--cluster-priority", dest="cluster_priority",
                               type=int,
                               help="set job priority on cluster.")
            group.add_argument("--cluster-queue", dest="cluster_queue",
                               type=str,
                               help="set cluster queue.")
            group.add_argument("--cluster-num-jobs", dest="cluster_num_jobs",
                               type=int,
                               help="number of jobs to submit to the queue execute "
                               "in parallel.")
            group.add_argument("--cluster-parallel",
                               dest="cluster_parallel_environment",
                               type=str,
                               help="name of the parallel environment to use ")
            group.add_argument("--cluster-options", dest="cluster_options",
                               type=str,
                               help="additional options for cluster jobs, passed "
                               "on to queuing system.")
            group.add_argument("--cluster-queue-manager",
                               dest="cluster_queue_manager",
                               type=str,
                               choices=["sge", "slurm", "torque", "pbspro"],
                               help="cluster queuing system ")
            group.add_argument("--cluster-memory-resource",
                               dest="cluster_memory_resource",
                               type=str,
                               help="resource name to allocate memory with ")
            group.add_argument("--cluster-memory-default",
                               dest="cluster_memory_default",
                               type=str,
                               help="default amount of memory to allocate ")

            parser.set_defaults(without_cluster=False,
                                cluster_queue=None,
                                cluster_priority=None,
                                cluster_num_jobs=None,
                                cluster_parallel_environment=None,
                                cluster_options=None,
                                cluster_memory_resource=None,
                                cluster_memory_default="unlimited",
                                cluster_queue_manager="sge")

        if add_output_options or add_pipe_options:
            group = parser.add_argument_group("Input/output options")

            if add_output_options:
                group.add_argument(
                    "-P", "--output-filename-pattern",
                    dest="output_filename_pattern", type=str,
                    help="OUTPUT filename pattern for various methods ")

                group.add_argument("-F", "--force-output", dest="output_force",
                                   action="store_true",
                                   help="force over-writing of existing files.")

                parser.set_defaults(output_filename_pattern="%s",
                                    output_force=False)

            if add_pipe_options:

                group.add_argument("-I", "--stdin", dest="stdin", type=str,
                                   help="file to read stdin from.")
                group.add_argument("-L", "--log", dest="stdlog", type=str,
                                   help="file with logging information.")
                group.add_argument("-E", "--error", dest="stderr", type=str,
                                   help="file with error information.")
                group.add_argument("-S", "--stdout", dest="stdout", type=str,
                                   help="file where output is to go. ")

                parser.set_defaults(stderr=sys.stderr)
                parser.set_defaults(stdout=sys.stdout)
                parser.set_defaults(stdlog=sys.stdout)
                parser.set_defaults(stdin=sys.stdin)

        if add_database_options:
            group = parser.add_argument_group("database connection options")
            group.add_argument(
                "--database-url", dest="database_url", type=str,
                help="database connection url, for example sqlite:///./csvdb.")

            group.add_argument(
                "--database-schema", dest="database_schema", type=str,
                help="database schem")

            parser.set_defaults(database_url="sqlite:///./csvdb")
            parser.set_defaults(database_schema=None)

        if return_parser:
            return parser

        if not no_parsing:
            global_args, unknown = parser.parse_known_args(argv[1:])

        if unknowns:
            opt1, opt2 = global_args, unknown
            retval = (opt1, opt2)
        else:
            opt1, opt2 = global_args, None
            retval = opt1

    _setup_seed(opt1)
    _setup_pipes(opt1, add_pipe_options)
    _setup_tracing(opt1)
    logger = _setup_logging(opt1, logger_callback)

    logger.info(get_header())
    logger.info(get_params(opt1))

    return retval


def stop(logger=None):
    """stop the experiment.

    This method performs final book-keeping, closes the output streams
    and writes the final log messages indicating script completion.
    """

    args = get_args()

    if args.loglevel >= 1 and global_benchmark:
        t = time.time() - global_starting_time
        args.stdlog.write(
            "######### Time spent in benchmarked functions #########\n")
        args.stdlog.write("# function\tseconds\tpercent\n")
        for key, value in list(global_benchmark.items()):
            args.stdlog.write(
                "# %s\t%6i\t%5.2f%%\n" % (key, value,
                                          (100.0 * float(value) / t)))
        args.stdlog.write(
            "#######################################################\n")

    if logger is None:
        logger = logging.getLogger("cgatcore")
    logger.info(get_footer())

    # close files
    if args.stdout != sys.stdout:
        args.stdout.close()
    # do not close log, otherwise error occurs in atext.py
    # if global_options.stdlog != sys.stdout:
    #   global_options.stdlog.close()

    if args.stderr != sys.stderr:
        args.stderr.close()

    if args.timeit_file:

        outfile = open(args.timeit_file, "a")

        if args.timeit_header:
            outfile.write("\t".join(
                ("name", "wall", "user", "sys", "cuser", "csys",
                 "host", "system", "release", "machine",
                 "start", "end", "path", "cmd")) + "\n")

        csystem, host, release, version, machine = list(map(str, os.uname()))
        uusr, usys, c_usr, c_sys = ["%5.2f" % x for x in os.times()[:4]]
        t_end = time.time()
        c_wall = "%5.2f" % (t_end - global_starting_time)

        if sys.argv[0] == "run.py":
            cmd = args[0]
            if len(args) > 1:
                cmd += " '" + "' '".join(args[1:]) + "'"
        else:
            cmd = sys.argv[0]

        result = "\t".join((global_options.timeit_name,
                            c_wall, uusr, usys, c_usr, c_sys,
                            host, csystem, release, machine,
                            time.asctime(time.localtime(global_starting_time)),
                            time.asctime(time.localtime(t_end)),
                            os.path.abspath(os.getcwd()),
                            cmd)) + "\n"

        outfile.write(result)
        outfile.close()


def get_output_file(section):
    '''return filename to write to, replacing any ``%s`` with section in
    the output pattern for files (``--output-filename-pattern``).
    '''
    return re.sub("%s", section, get_args().output_filename_pattern)


def open_output_file(section, mode="w", encoding="utf-8"):
    """open file for writing substituting section in the
    output_pattern (if defined).

    This method will automatically create any parent directories that
    are missing.

    If the filename ends with ".gz", the output is opened as a gzip'ed
    file.

    Arguments
    ---------
    section : string
       section will replace any %s in the pattern for output files.

    mode : char
       file opening mode

    Returns
    -------
    File
        an opened file

    """

    fn = get_output_file(section)
    args = get_args()

    if fn == "-":
        return args.stdout

    if not args.output_force and os.path.exists(fn):
        raise OSError(
            "file %s already exists, use --force-output to "
            "overwrite existing files.".format(fn))

    return open_file(fn, mode=mode, create_dir=True, encoding=encoding)


def run(statement,
        return_stdout=False,
        return_stderr=False,
        return_popen=False,
        on_error="raise",
        encoding="utf-8",
        **kwargs):
    '''execute a command line statement.

    By default this method returns the code returned by the executed
    command. If *return_stdout* is True, the contents of stdout are
    returned as a string, likewise for *return_stderr*.

    If *return_popen*, the Popen object is returned.

    ``kwargs`` are passed on to subprocess.call,
    subprocess.check_output or subprocess.Popen.

    Arguments
    ---------

    on_error: string
       Action to perform on error. Valid actions are "ignore" and "raise".

    Raises
    ------

    OSError
       If process failed or was terminated.

    '''

    # remove new lines
    statement = " ".join(re.sub("\t+", " ", statement).split("\n")).strip()

    if "<(" in statement:
        shell = os.environ.get('SHELL', "/bin/bash")
        if "bash" not in shell:
            raise ValueError(
                "require bash for advanced shell syntax: <()")
        # Note: pipes.quote is deprecated. In Py3, use shlex.quote
        # (not present in Py2.7)
        statement = "%s -c %s" % (shell, pipes.quote(statement))

    if return_stdout:
        try:
            output = subprocess.check_output(statement, shell=True, **kwargs)
        except subprocess.CalledProcessError as e:
            if on_error == "raise":
                raise
            output = e.output
        return output.decode(encoding)

    elif return_stderr:
        # expect that process fails
        p = subprocess.Popen(statement,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             **kwargs)
        stdout, stderr = p.communicate()
        return stderr.decode(encoding)

    elif return_popen:
        return subprocess.Popen(statement, shell=True, **kwargs)

    else:
        retcode = subprocess.call(statement, shell=True, **kwargs)
        if retcode < 0:
            raise OSError(
                "process was terminated by signal: {}".format(-retcode))
        elif retcode > 0:
            raise OSError("process exited with code: {}".format(retcode))
        return retcode


# some convenient decorators
def benchmark(func):
    """decorator collecting wall clock time spent in decorated method."""

    def wrapper(*arg):
        t1 = time.time()
        res = func(*arg)
        t2 = time.time()
        key = "%s:%i" % (func.__name__, func.__code__.co_firstlineno)
        global_benchmark[key] += t2 - t1
        global_options.stdlog.write(
            '## benchmark: %s completed in %6.4f s\n' % (key, (t2 - t1)))
        global_options.stdlog.flush()
        return res
    return wrapper


class cached_function(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.

    Taken from http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = (args, frozenset(list(kwargs.items())))
        try:
            return self.cache[key]
        except KeyError:
            value = self.func(*args, **kwargs)
            self.cache[key] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kwargs)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)

    def delete(self, *args, **kwargs):
        """remove a cache entry"""
        key = (args, frozenset(list(kwargs.items())))
        del self.cache[key]


class cached_property(object):
    '''Decorator for read-only properties.

    Modified from https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    '''

    def __init__(self, hello):
        print(hello)

    def __call__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        return self

    def __get__(self, inst, owner):
        try:
            value = inst._cache[self.__name__]
        except KeyError:
            value = self.fget(inst)
            try:
                cache = inst._cache
            except AttributeError:
                cache = inst._cache = {}
            cache[self.__name__] = value
        return value


class cached_method(object):

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        if instance is None:
            return self.func
        return functools.partial(self, instance)

    def __call__(self, *args, **kwargs):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}

        key = (self.func, args[1:], frozenset(list(kwargs.items())))
        try:
            return cache[key]
        except KeyError:
            value = self.func(*args, **kwargs)
            cache[key] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kwargs)


class Counter(object):
    '''a counter class.

    The counter acts both as a dictionary and a object permitting
    attribute access.

    Counts are automatically initialized to 0.

    Instantiate and use like this::

       c = Counter()
       c.input += 1
       c.output += 2
       c["skipped"] += 1

       print str(c)
    '''

    __slots__ = ["_counts"]

    def __init__(self):
        """Store data returned by function."""
        object.__setattr__(self, "_counts", collections.defaultdict(int))

    def __setitem__(self, key, value):
        self._counts[key] = value

    def __getitem__(self, key):
        return self._counts[key]

    def __str__(self):
        return ", ".join("%s=%i" % x for x in self._counts.items())

    def items(self):
        return self._counts.items()

    def __iadd__(self, other):
        for key, val in other.items():
            self._counts[key] += val
        return self

    def iteritems(self):
        return iter(self._counts.items())

    def values(self):
        return self._counts.values()

    def asTable(self, as_rows=True):
        '''return values as tab-separated table (without header).

        Key, value pairs are sorted lexicographically.
        '''
        if as_rows:
            return '\n'.join("%s\t%i" % x
                             for x in sorted(self._counts.items()))
        else:
            columns, values = list(zip(*sorted(self._counts.items())))
            return '{}\n{}'.format("\t".join(columns),
                                   "\t".join(map(str, values)))

    def __getattr__(self, name):
        return self._counts[name]

    def __setattr__(self, name, value):
        self._counts[name] = value


def trace_calls(frame, event, arg):
    """trace functions calls for debugging purposes.

    See https://pymotw.com/2/sys/tracing.html
    """
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    if caller is None:
        print('# called: %s:%s (%s) (from None' %
              (func_name, func_line_no, func_filename))
    else:
        caller_line_no = caller.f_lineno
        caller_filename = caller.f_code.co_filename
        print('# called: %s:%s (%s) (from %s:%s)' %
              (func_name, func_line_no, func_filename,
               caller_filename, caller_line_no))
    return


def enable_tracing(level="function"):
    if level == "function":
        sys.settrace(trace_calls)



# This implimentation was inspired/copied in parts from the FTP in smakemake remotes/FTP.py
import os
import sys
import ftplib
from contextlib import contextmanager

try:
    import ftputil
    import ftputil.session
except ImportError as e:
    raise WorkflowError(
        "The ftputil package needs to be installed. %s" % (e.msg))

from cgatcore.remote import AbstractRemoteObject


class FTPRemoteObject(AbstractRemoteObject):
    '''This is a class that will interact with an a secure ftp server.'''

    def __init__(self, *args, stay_on_remote=False, immediate_close=False, **kwargs):
        super(FTPRemoteObject, self).__init__(
            *args, stay_on_remote=stay_on_remote, **kwargs)

    @contextmanager
    def ftpc(self):

        if (not hasattr(self, "conn") or
            (hasattr(self, "conn") and
             not isinstance(self.conn, ftputil.FTPHost))) or self.immediate_close:
            args_use = self.provider.args
            if len(self.args):
                args_use = self.args

                kwargs_use = {}
                kwargs_use['host'] = self.host
                kwargs_use['password'] = None
                kwargs_use['username'] = None
                kwargs_use['port'] = int(self.port) if self.port else 21
                kwargs_use['encrypt_data_channel'] = self.encrypt_data_channel

                for k, v in self.provider.kwargs.items():
                    kwargs_use[k] = v
                for k, v in self.kwargs.items():
                    kwargs_use[k] = v

                ftp_base_class = ftplib.FTP_LTS if kwargs_use['encrypt_data_channel'] else ftplib.FTP

                ftp_session_factory = ftputil.session.session_factory(
                    base_class=ftp_base_class,
                    port=kwargs_use['port'],
                    encrypt_data_channel=kwargs_use['encrypt_data_channel'],
                    debug_level=None)

                conn = ftputil.FTPHost(
                    kwargs_use['host'],
                    kwargs_use['username'],
                    kwargs_use['password'],
                    session_factory=ftp_session_factory)

                if self.immediate_close:
                    yield conn
                else:
                    self.conn = conn
                    yield self.conn
            elif not self.immediate_close:
                yield self.conn

        if self.immediate_close:
            try:
                conn.seep_alive()
                conn.close()
            except Exception:
                pass

    def exists(self):
        if self._matched_address:
            with self.ftpc as ftpc:
                return ftpc.path.exists(self.remote_path)
            return False
        else:
            raise FTPFileException(
                "The file cannot be parsed as an TFP path in form 'host:port/path/to/file': %s" % self.local_file())

    def download(self, make_dest_dir=True):
        with self.ftpc() as ftpc:
            if self.exists():
                # if the dest path does not exist
                if make_dest_dirs:
                    os.makedirs(os.path.dirname(
                        self.local_path, exists_ok=True))
                try:
                    ftpc.syncronize_times()
                except Exception:
                    pass

                ftpc.download(source=self.remote_path, target=self.local_path)
                os.sync()
            else:
                raise FTPFileException(
                    "The file does not exist remotely: %s" % self.local_file())

    def upload(self):
        with self.ftpc() as ftpc:
            ftpc.synchronize_times()
            ftpc.upload(source=self.local_path, target=self.remote_path)

    def delete_file(self):
        raise NotImplementedError("Cannot delete files from an FTP server")



'''cluster.py - cluster utility functions for ruffus pipelines
==============================================================

This module abstracts the DRMAA native specification and provides
convenience functions for running Drmaa jobs.

Currently SGE, SLURM, Torque and PBSPro are supported.

Reference
---------

'''

import re
import math
import collections
import os
import stat
import time
import datetime
import logging
import gevent
import cgatcore.experiment as E

try:
    import drmaa
except (ImportError, RuntimeError, OSError):
    pass


def get_logger():
    return logging.getLogger("cgatcore.pipeline")


JobInfo = collections.namedtuple("JobInfo", ("jobId", "resourceUsage"))


# Timeouts for event loop
GEVENT_TIMEOUT_SACCT = 5
GEVENT_TIMEOUT_WAIT = 1


class DRMAACluster(object):

    # dictionary mapping resource usage fields returned by DRMAA
    # to a common set of names.
    map_drmaa2benchmark_data = {}

    def __init__(self, session, ignore_errors=False):
        self.session = session
        self.ignore_errors = ignore_errors

    def get_resource_usage(self, job_id, retval, hostname):
        retval.resourceUsage["hostname"] = hostname
        return [retval]

    def setup_drmaa_job_template(self,
                                 drmaa_session,
                                 job_name,
                                 job_memory,
                                 job_threads,
                                 working_directory,
                                 **kwargs):
        '''Sets up a Drmma job template. Currently SGE, SLURM, Torque and PBSPro are
        supported'''
        if not job_memory:
            raise ValueError("Job memory must be specified when running"
                             "DRMAA jobs")

        jt = drmaa_session.createJobTemplate()
        jt.workingDirectory = working_directory
        jt.jobEnvironment = {'BASH_ENV': '~/.bashrc'}
        jt.args = []
        if not re.match("[a-zA-Z]", job_name[0]):
            job_name = "_" + job_name

        spec = self.get_native_specification(job_name,
                                             job_memory,
                                             job_threads,
                                             **kwargs)

        jt.nativeSpecification = " ".join(spec)

        # keep stdout and stderr separate
        jt.joinFiles = False

        self.update_template(jt)
        return jt

    def update_template(self, jt):
        pass

    def collect_single_job_from_cluster(self,
                                        job_id,
                                        statement,
                                        stdout_path, stderr_path,
                                        job_path):
        '''collects a single job on the cluster.

        This method waits until a job has completed and returns
        stdout, stderr and resource usage.
        '''
        try:
            retval = self.session.wait(
                job_id, drmaa.Session.TIMEOUT_WAIT_FOREVER)
        except Exception as msg:
            # ignore message 24, indicates jobs that have been qdel'ed
            if not str(msg).startswith("code 24"):
                raise
            retval = None

        stdout, stderr = self.get_drmaa_job_stdout_stderr(
            stdout_path, stderr_path)

        if retval is not None:
            error_msg = None
            if retval.exitStatus == 0:
                if retval.wasAborted is True:
                    error_msg = (
                        "Job {} has exit status 0, but marked as hasAborted=True, hasExited={} "
                        "(Job may have been cancelled by the user or the scheduler due to memory constraints)"
                        "The stderr was \n{}\nstatement = {}".format(
                            job_id, retval.hasExited, "".join(stderr), statement))
                if retval.hasSignal is True:
                    error_msg = ("Job {} has zero exitStatus {} but received signal: hasExited={},  wasAborted={}"
                                 "hasSignal={}, terminatedSignal='{}' "
                                 "\nstatement = {}".format(
                                     job_id, retval.exitStatus, retval.hasExited, retval.wasAborted,
                                     retval.hasSignal, retval.terminatedSignal,
                                     statement))
            else:
                error_msg = ("Job {} has non-zero exitStatus {}: hasExited={},  wasAborted={}"
                             "hasSignal={}, terminatedSignal='{}' "
                             "\nstatement = {}".format(
                                 job_id, retval.exitStatus, retval.hasExited, retval.wasAborted,
                                 retval.hasSignal, retval.terminatedSignal,
                                 statement))

            if error_msg:
                if stderr:
                    error_msg += "\n stderr = {}".format("".join(stderr))
                if self.ignore_errors:
                    get_logger().warning(error_msg)
                else:
                    raise OSError(error_msg)
        else:
            retval = JobInfo(job_id, {})

        # get hostname from job script
        try:
            hostname = stdout[-3][:-1]
        except IndexError:
            hostname = "unknown"

        try:
            resource_usage = self.get_resource_usage(job_id, retval, hostname)
        except (ValueError, KeyError, TypeError, IndexError) as ex:
            E.warn("could not collect resource usage for job {}: {}".format(job_id, ex))
            retval.resourceUsage["hostname"] = hostname
            resource_usage = [retval]

        try:
            os.unlink(job_path)
        except OSError:
            self.logger.warn(
                ("temporary job file %s not present for "
                 "clean-up - ignored") % job_path)

        return stdout, stderr, resource_usage

    def get_drmaa_job_stdout_stderr(self, stdout_path, stderr_path,
                                    tries=5, encoding="utf-8"):
        '''get stdout/stderr allowing for some lag.

        Try at most *tries* times. If unsuccessfull, throw OSError

        Removes the files once they are read.

        Returns tuple of stdout and stderr as unicode strings.
        '''
        x = tries
        while x >= 0:
            if os.path.exists(stdout_path):
                break
            gevent.sleep(GEVENT_TIMEOUT_WAIT)
            x -= 1

        x = tries
        while x >= 0:
            if os.path.exists(stderr_path):
                break
            gevent.sleep(GEVENT_TIMEOUT_WAIT)
            x -= 1

        try:
            with open(stdout_path, "r", encoding=encoding) as inf:
                stdout = inf.readlines()
        except IOError as msg:
            get_logger().warning("could not open stdout: %s" % msg)
            stdout = []

        try:
            with open(stderr_path, "r", encoding=encoding) as inf:
                stderr = inf.readlines()
        except IOError as msg:
            get_logger().warning("could not open stdout: %s" % msg)
            stderr = []

        try:
            os.unlink(stdout_path)
            os.unlink(stderr_path)
        except OSError as msg:
            pass

        return stdout, stderr

    def set_drmaa_job_paths(self, job_template, job_path):
        '''Adds the job_path, stdout_path and stderr_paths
           to the job_template.
        '''
        job_path = os.path.abspath(job_path)
        os.chmod(job_path, stat.S_IRWXG | stat.S_IRWXU)

        stdout_path = job_path + ".stdout"
        stderr_path = job_path + ".stderr"

        job_template.remoteCommand = job_path
        job_template.outputPath = ":" + stdout_path
        job_template.errorPath = ":" + stderr_path

        return stdout_path, stderr_path

    def map_resource_usage(self, resource_usage, data2type):
        """return job metrics mapped to common name and converted to right type."""
        def _convert(key, v, tpe):
            if v is None:
                return None
            else:
                try:
                    return tpe(v)
                except ValueError as ex:
                    E.warning("could not convert {} with value '{}' to {}: {}".format(
                        key, v, tpe, ex))
                    return v

        return dict([(key,
                      _convert(key, resource_usage.get(self.map_drmaa2benchmark_data.get(key, key), None), tpe))
                     for key, tpe in data2type.items()])


class SGECluster(DRMAACluster):

    map_drmaa2benchmark_data = {
        "wall_t": "ru_wallclock",
        "cpu_t": "cpu",
        "user_t": "ru_utime",
        "sys_t": "ru_stime",
        # "child_user_t":
        # "child_sys_t",
        "shared_data": "ru_ixrss",
        "io_input": "ru_inblock",
        "io_output": "ru_oublock",
        # SGE this is integral memory usage (memory * time)
        "average_memory_total": "mem",
        # needs to be set manually as user + system times divided by the total running time
        "percent_cpu": "percent_cpu",
        # SGE this is integral memory usage
        "average_rss": "mem",
        "max_rss": "ru_maxrss",
        "max_vmem": "maxvmem",
        "minor_page_faults": "ru_minflt",
        "swapped": "ru_nswap",
        "context_switches_involuntarily": "ru_nvcsw",
        "context_switches_voluntarily": "ru_nivcsw",
        "average_uss": "ru_isrss",
        "signal": "signal",
        "socket_received": "ru_msgrcv",
        "socket_sent": "ru_msgsnd",
        "major_page_faults": "ru_majflt",
        "unshared_data": "ru_idrss"}

    def get_native_specification(self,
                                 job_name,
                                 job_memory,
                                 job_threads,
                                 **kwargs):
        # see: ? cannot find documentation on the SGE native spec
        spec = ["-V",
                "-notify",  # required for signal handling
                "-N %s" % job_name]

        spec.append("-p {}".format(kwargs.get("priority", 0)))

        spec.append(kwargs.get("options", ""))

        if not kwargs["memory_resource"]:
            raise ValueError("cluster memory resource not specified")

        if job_memory != "unlimited":
            for resource in kwargs.get("memory_resource", "").split(","):
                spec.append("-l {}={}".format(resource, job_memory))

        if job_threads > 1:
            if not kwargs["parallel_environment"]:
                raise ValueError("parallel environment not specified")
            spec.append("-pe {} {} -R y".format(
                kwargs.get("parallel_environment", "smp"), job_threads))

        if "pe_queue" in kwargs and job_threads > 1:
            spec.append("-q {}".format(kwargs["pe_queue"]))
        elif kwargs['queue'] != "NONE":
            spec.append("-q {}".format(kwargs["queue"]))

        return spec


class SlurmCluster(DRMAACluster):

    map_drmaa2benchmark_data = {
        "hostname": "NodeList",
        "job_id": "JobID",
        "submission_time": "Submit",
        "start_time": "Start",
        "end_time": "End",
        "slots": "NCPUS",
        "exit_status": "ExitCode",
        "wall_t": "ElapsedRaw",
        "cpu_t": "CPUTimeRaw",
        "user_t": "UserCPU",
        "sys_t": "SystemCPU",
        "io_input": "MaxDiskRead",
        "io_output": "MaxDiskWrite",
        "average_memory_total": "AveVMSize",
        "average_rss": "AveRSS",
        "max_rss": "MaxRSS",
        "max_vmem": "MaxVMSize",
        "minor_page_faults": "AvePages",
        "average_uss": "AveRSS",
        "signal": "DerivedExitCode",
        "major_page_faults": "MaxPages"}

    def get_native_specification(self,
                                 job_name,
                                 job_memory,
                                 job_threads,
                                 **kwargs):
        # SLURM DOCS:
        # http://apps.man.poznan.pl/trac/slurm-drmaa
        # https://computing.llnl.gov/linux/slurm/cons_res_share.html
        #
        # The SLURM Consumable Resource plugin is required
        # The "CR_CPU_Memory" resource must be specified
        #
        # i.e. in slurm.conf:
        # SelectType=select/cons_res
        # SelectTypeParameters=CR_CPU_Memory
        #
        # * Note that --cpus-per-task will actually refer to cores
        #   with the appropriate Node configuration
        #
        # SLURM-DRMAA DOCS - Note that version 1.2 (SVN) is required
        # http://apps.man.poznan.pl/trac/slurm-drmaa
        #
        # Not implemented:
        # -V: SLURM automatically passess the environment variables
        # -p: does not appear to be part of the slurm drmaa native spec
        #
        # TODO: add "--account" (not sure the best way to fill param).

        spec = ["-J %s" % job_name]

        spec.append(kwargs.get("options", ""))
        spec.append("--cpus-per-task={}".format(job_threads))

        # Note the that the specified memory must be per CPU
        # for consistency with the implemented SGE approach

        if job_memory != "unlimited":
            if job_memory.endswith("G"):
                job_memory_per_cpu = int(
                    math.ceil(float(job_memory[:-1]) * 1000))
            elif job_memory.endswith("M"):
                job_memory_per_cpu = int(math.ceil(float(job_memory[:-1])))
            else:
                raise ValueError('job memory unit not recognised for SLURM, '
                                 'must be either "M" (for Mb) or "G" (for Gb),'
                                 ' e.g. 1G or 1000M for 1 Gigabyte of memory')

            spec.append("--mem-per-cpu={}".format(job_memory_per_cpu))

        # set the partition to use (equivalent of SGE queue)
        spec.append("--partition={}".format(kwargs["queue"]))

        return spec

    @classmethod
    def parse_accounting_data(cls, data, retval):

        def convert_value(*pair):
            k, v = pair[0]
            if k in ("UserCPU", "SystemCPU"):
                n = 0
                for x, f in zip(re.split("[-:]", v)[::-1], (1, 60, 3600, 86400)):
                    n += float(x) * f
                v = n
            elif k in ("Start", "End", "Submit"):
                try:
                    v = time.mktime(datetime.datetime.strptime(
                        v, "%Y-%m-%dT%H:%M:%S").timetuple())
                except ValueError as ex:
                    E.warning("could not parse time value '{}' for field '{}': {}".format(
                        v, k, ex))
                    v = None
            elif k in ("ExitCode", ):
                v = int(v.split(":")[0])
            elif v.endswith("K"):
                v = float(v[:-1]) * 1000
            elif v.endswith("M"):
                v = float(v[:-1]) * 1000000
            elif v.endswith("G"):
                v = float(v[:-1]) * 1000000000
            try:
                v = int(v)
            except ValueError:
                pass
            return k, v

        d = dict(map(convert_value, zip(
            cls.map_drmaa2benchmark_data.values(), data.split("|"))))
        retval = retval._replace(resourceUsage=d)
        return [retval]

    def get_resource_usage(self, job_id, retval, hostname):
        # delay to help with sync'ing of book-keeping
        gevent.sleep(GEVENT_TIMEOUT_SACCT)
        statement = "sacct --noheader --units=K --parsable2 --format={} -j {} ".format(
            ",".join(self.map_drmaa2benchmark_data.values()), job_id)

        stdout = E.run(statement, return_stdout=True).splitlines()
        if len(stdout) != 2:
            E.warn("expected 2 lines in {}, but got {}".format(
                statement, len(stdout)))

        return self.parse_accounting_data(stdout[-1], retval)


class TorqueCluster(DRMAACluster):

    def get_native_specification(self,
                                 job_name,
                                 job_memory,
                                 job_threads,
                                 **kwargs):
        # PBS Torque native specifictation:
        # http://apps.man.poznan.pl/trac/pbs-drmaa

        spec = ["-N {}".format(job_name), ]

        # again, I don't know if mem is same across all sites, or just a
        # common default, so allow to be set via memory_resource, with "mem"
        # default (for backwards compatibility).

        resource_requests = list()

        if job_memory != "unlimited":
            for resource in kwargs.get("memory_resource", "mem").split(","):
                resource_requests.append("{}={}".format(resource, job_memory))

        if job_threads > 1:
            # don't know if this is standard resource names or if
            # it varies site to site. For now, I will assume "nodes" is
            # standard and name of resource for multiple CPUs is stored in
            # "parrellel environment"
            pe = kwargs.get("parallel_environment", "ppn")
            resource_requests.append("nodes=1")
            resource_requests.append("{}={}".format(pe, job_threads))

        if resource_requests:
            spec.append("-l " + ":".join(resource_requests))

        if "pe_queue" in kwargs and job_threads > 1:
            spec.append("-q {}".format(kwargs["pe_queue"]))
        elif kwargs['queue'] != "NONE":
            spec.append("-q {}".format(kwargs["queue"]))

        spec.append(kwargs.get("options", ""))

        return spec

    def update_template(self, jt):
        # There is no equivalent to sge -V option for pbs-drmaa
        # recreating this...
        jt.jobEnvironment = os.environ
        jt.jobEnvironment.update({'BASH_ENV': os.path.join(os.path.expanduser("~"),
                                                           '.bashrc')})


class PBSProCluster(DRMAACluster):

    def get_native_specification(self,
                                 job_name,
                                 job_memory,
                                 job_threads,
                                 **kwargs):

        # PBS Pro docs
        # http://www.pbsworks.com/PBSProduct.aspx?n=PBS-Professional&c=Overview-and-Capabilities
        # http://technion.ac.il/usg/tamnun/PBSProUserGuide12.1.pdf

        # DRMAA for PBS Pro is the same as for torque:
        # http://apps.man.poznan.pl/trac/pbs-drmaa
        # Webpages with some examples:
        # https://wiki.galaxyproject.org/Admin/Config/Performance/cluster#PBS
        # https://sites.google.com/a/case.edu/hpc-upgraded-cluster/home/Software-Guide/pbs-drmaa
        # https://albertsk.files.wordpress.com/2011/12/pbs.pdf

        # PBS Pro has some differences with torque so separating

        # Set environment variables in .bashrc:
        # PBS_DRMAA_CONF to eg ~/.pbs_drmaa.conf
        # DRMAA_LIBRARY_PATH to eg /xxx/libdrmaa.so

        # PBSPro only takes the first 15 characters, throws
        # uninformative error if longer.  mem is maximum amount of RAM
        # used by job; mem_free doesn't seem to be available.
        spec = ["-N {}".format(job_name[0:15]), ]

        if "mem" not in kwargs["options"]:
            spec.append("-l mem={}".format(job_memory))

        # Leaving walltime to be specified by user as difficult to set
        # dynamically and depends on site/admin configuration of
        # default values. Likely means setting for longest job with
        # trade-off of longer waiting times for resources to be
        # available for other jobs.

        if job_threads > 1:
            # TO DO 'select=1' determines de number of nodes. Should
            # go in a config file.  mem is per node and maximum memory
            # Site dependent but in general setting '#PBS -l
            # select=NN:ncpus=NN:mem=NN{gb|mb}' is sufficient for
            # parallel jobs (OpenMP, MPI).  Also architecture
            # dependent, jobs could be hanging if resource doesn't
            # exist.
            # mem outside of -l select errors (ie passing -l mem=4G errors)
            # TO DO: Kill if long waiting time?
            spec.append("-l select=1:ncpus=%s:mem=%s".format(job_threads,
                                                             job_memory
                                                             )
                        )

        if "pe_queue" in kwargs and job_threads > 1:
            spec.append("-q {}".format(kwargs["pe_queue"]))
        elif kwargs['queue'] != "NONE":
            spec.append("-q {}".format(kwargs["queue"]))

        spec.append(kwargs.get("options", ""))
        return spec

    def update_template(self, jt):
        # The directive #PBS -V exists and works in a qsub script but errors here
        # so using the following as for torque:
        jt.jobEnvironment = os.environ
        jt.jobEnvironment.update(
            {'BASH_ENV': os.path.join(os.path.expanduser("~"), '.bashrc')})


def get_queue_manager(queue_manager, *args, **kwargs):

    qm = queue_manager.lower()

    if qm == "sge":
        return SGECluster(*args, **kwargs)
    elif qm == "slurm":
        return SlurmCluster(*args, **kwargs)
    elif qm == "torque":
        return TorqueCluster(*args, **kwargs)
    elif qm == "pbspro":
        return PBSProCluster(*args, **kwargs)
    else:
        raise ValueError(
            "Queue manager {} not supported".format(queue_manager))


MAP_QACCT2BENCHMARK_DATA = {
    "task": "task",
    "statement": "statement",
    "hostname": "hostname",
    "started": "start_time",
    "completed": "end_time",
    "total_t": "total_t",
    "wall_t": "cpu",
    "user_t": "ru_utime",
    "sys_t": "ru_stime",
    # "child_user_t",
    # "child_sys_t",
    "shared_data": "ru_ixrss",
    "io_input": "ru_inblock",
    "io_output": "ru_oublock",
    # SGE this is integral memory usage
    "average_memory_total": "mem",
    "percent_cpu": "cpu",
    # SGE this is integral memory usage
    "average_rss": "mem",
    "max_rss": "ru_maxrss",
    "minor_page_faults": "ru_minflt",
    "swapped": "ru_nswap",
    "context_switches_involuntarily": "ru_nvcsw",
    "context_switches_voluntarily": "ru_nivcsw",
    "average_uss": "ru_isrss"
}



# This implimentation was inspired by the SFTP in smakemake remotes repo
import os
import sys
from contextlib import contextmanager

try:
    import pysftp
except ImportError as e:
    raise WorkflowError(
        "The pysftp package needs to be installed. %s" % (e.msg))

from cgatcore.remote import AbstractRemoteObject


class SFTPRemoteObject(AbstractRemoteObject):
    '''This is a class that will interact with an a secure ftp server.'''

    def __init__(self, *args, keep_local=False, provider=None, **kwargs):
        super(SFTPRemoteObject, self).__init__(
            *args, keep_local=keep_local, provider=provider, **kwargs)

    @contextmanager
    def sftpc(self):

        args_use = self.provider.args
        if len(self.args):
            args_use = self.args

        kwargs_use = {}
        kwargs_use['host'] = self.host
        kwargs_use['port'] = int(self.port) if self.port else 22
        for k, v in self.provider.kwargs.items():
            kwargs_use[k] = v
        for k, v in self.kwargs.items():
            kwargs_use[k] = v

        conn = pysftp.Connection(*args_use, **kwargs_use)
        yield conn
        conn.close()

    def exists(self):
        if self._matched_address:
            with self.sftpc as sftpc:
                return sftpc.exists(self.remote_path)
                if sftpc.exists(self.remote_path):
                    return sftpc.isfile(self.remote_path)
                return False
        else:
            raise SFTPFileException(
                "The file cannot be parsed as an STFP path in form 'host:port/path/to/file': %s" % self.local_file())

    def download(self, make_dest_dir=True):
        with self.sftpc() as sftpc:
            if self.exists():
                # if the dest path does not exist
                if make_dest_dirs:
                    os.makedirs(os.path.dirname(
                        self.local_path, exists_ok=True))

                sftpc.get(remotepath=self.remote_path,
                          localpath=self.local_path, preserve_mtime=True)
                os.sync()
            else:
                raise SFTPFileException(
                    "The file cannot be parsed as an STFP path in "
                    "form 'host:port/path/to/file': %s" % self.local_file())

    def upload(self):
        with self.sftpc() as sftpc:
            sftpc.put(localpath=self.local_path, remotepath=self.remote_path,
                      confirm=True, preserve_mtime=True)

    def delete_file(self):
        raise NotImplementedError("Cannot delete files from an SFTP server")



class EmptyRunner(object):

    def __init__(self, name):
        self.__name__ = name

    def __call__(self, *args, **kwargs):
        pass


class PassThroughRunner(object):
    def __init__(self, name, f):
        self.__name__ = name
        self.f = f

    def __call__(self, *args, **kwargs):
        self.f(*args, **kwargs)



import os
import sys

try:
    import google.cloud
    from google.cloud import storage
except ImportError as e:
    raise WorkFlowError("The google cloud SDK needs to be installed. %s"
                        % e.msg)

from cgatcore.remote import AbstractRemoteObject


class GCRemoteObject(AbstractRemoteObject):
    '''This is a class that will interact with an Google cloud object store.'''

    def __init__(self, *args, **kwargs):
        super(GCRemoteObject, self).__init__(*args, **kwargs)

        self._GCobject = GCConnection(*args, **kwargs)

    def exists(self, bucket):
        try:
            self._S3object.bucket_exists(bucket_name)
        except Exception:
            raise KeyError("The bucket does not exist in google storage")

    def download(self, bucket_name, key, file_dir):
        self._GCobject.remote_download(bucket_name, key, file_dir)
        os.sync()
        return file_dir

    def upload(self, bucket_name, key, file_dir):
        self._GCobject.remote_upload(bucket_name, file_dir, key)
        return file_dir

    def delete_file(self, bucket_name, key):
        self._GCobject.remote_delete_file(bucket_name, key)
        return key


class GCConnection():
    '''This is a connection to a remote google cloud bucket
       using the goggle.cloud SDK.'''

    def __init__(self, *args, **kwargs):
        #

        self.GC = storage.Client(*args, **kwargs)
        self.blob = None

    def bucket_exists(self, bucket_name):
        try:
            self.GC.get_bucket(bucket_name)
            return True
        except KeyError as e:
            print("%s bucket does not exist in google cloud" % (e))

    def remote_download(self,
                        bucket_name,
                        key,
                        dest_dir):
        '''Download data/file (blobs) from an google storage bucket.'''

        if not bucket_name:
            raise ValueError("Bucket name must be specified to download file")
        if not key:
            raise ValueError("Key must be specified to download file")

        if dest_dir:
            dest_path = os.path.realpath(os.path.expanduser(dest_dir))

        b = self.GC.get_bucket(bucket_name)
        blob = b.blob(key)

        try:
            blob.download_to_filename(dest_path)
        except Exception:
            raise Exception('''no file was downloaded, make sure the correct
                            file or path is specified. It currently is: {}'''.format(dest_path))

    def remote_upload(self,
                      bucket_name,
                      file_dir,
                      key=None):
        '''Upload data/file (blob) to a google cloud  bucket.'''

        file_path = os.path.realpath(os.path.expanduser(file_dir))

        if not bucket_name:
            raise ValueError("Bucket name must be specified to upload file")
        if not os.path.exists(file_dir):
            raise ValueError(
                "File path specified does not exitis: {}".format(file_path))
        if not os.path.isfile(file_dir):
            raise ValueError(
                "File path specified is not a file: {}".format(file_path))

        if not self.bucket_exists(bucket_name):
            self.GC.create_bucket(bucket_name)

        b = self.GC.get_bucket(bucket_name)
        blob = b.blob(key)

        try:
            blob.upload_from_filename(file_path)
        except Exception:
            raise Exception(
                "filename is not correctly specified: {}".format(file_dir))

    def remote_delete_file(self, bucket_name, key):
        '''Will remove the object from the remote S3 bucket'''

        if not bucket_name:
            raise ValueError("Bucket name must be specified to download file")
        if not key:
            raise ValueError("Key must be specified to download file")

        b = self.GC.get_bucket(bucket_name)
        blob = b.blob(key)
        f_delete = blob.delete()

        return key



import os
import sys

try:
    import boto3
    import botocore
except ImportError as e:
    raise WorkflowError(
        "The boto3 package needs to be installed. %s" % (e.msg))

from cgatcore.remote import AbstractRemoteObject


class S3RemoteObject(AbstractRemoteObject):
    '''This is a class that will interact with an AWS object store.'''

    def __init__(self, *args, **kwargs):
        super(S3RemoteObject, self).__init__(*args, **kwargs)

        self._S3object = S3Connection(*args, **kwargs)

    def exists(self, bucket_name):

        try:
            self._S3object.bucket_exists(bucket_name)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                return False
            else:
                raise S3FileException(
                    "The file cannot be parsed as an s3 path in form 'bucket/key': %s" % self.local_file())

        return True

    def download(self, bucket_name, key, file_dir):
        self._S3object.remote_download(bucket_name, key, file_dir)
        os.sync()  # ensure flush to disk
        return file_dir

    def upload(self, bucket_name, key, file_dir):
        self._S3object.remote_upload(bucket_name, file_dir, key)
        return file_dir

    def delete_file(self, bucket_name, key):
        self._S3object.remote_delete_file(bucket_name, key)
        return key


class S3Connection():
    '''This is a connection to a remote S3 bucket for AWS
    server using the boto3 API.'''

    def __init__(self, *args, **kwargs):
        #

        self.S3 = boto3.resource("s3", **kwargs)

    def bucket_exists(self, bucket_name):
        try:
            self.S3.meta.client.head_bucket(Bucket=bucket_name)
            return True
        except Exception:
            return False

    def remote_download(self,
                        bucket_name,
                        key,
                        dest_dir):
        '''Download data/file from an S3 bucket.'''

        if not bucket_name:
            raise ValueError("Bucket name must be specified to download file")
        if not key:
            raise ValueError("Key must be specified to download file")

        if dest_dir:
            dest_path = os.path.realpath(os.path.expanduser(dest_dir))

        f = self.S3.Object(bucket_name, key)

        try:
            f.download_file(dest_path)
        except Exception:
            raise Exception('''no file was downloaded, make sure the correct
                            file or path is specified. It currently is: {}'''.format(dest_path))

    def remote_upload(self,
                      bucket_name,
                      file_dir,
                      key=None):
        '''Upload data/file to an S3 bucket.'''

        file_path = os.path.realpath(os.path.expanduser(file_dir))

        if not bucket_name:
            raise ValueError("Bucket name must be specified to upload file")
        if not os.path.exists(file_dir):
            raise ValueError(
                "File path specified does not exitis: {}".format(file_path))
        if not os.path.isfile(file_dir):
            raise ValueError(
                "File path specified is not a file: {}".format(file_path))

        if not self.bucket_exists(bucket_name):
            # Implement other features fuch as CreateBucketConfiguration
            self.S3.create_bucket(Bucket=bucket_name)

        f = self.S3.Object(bucket_name, key)

        try:
            f.upload_file(file_path)
        except Exception:
            raise Exception(
                "filename is not correctly specified: {}".format(file_dir))

    def remote_delete_file(self, bucket_name, key):
        '''Will remove the object from the remote S3 bucket'''

        if not bucket_name:
            raise ValueError("Bucket name must be specified to download file")
        if not key:
            raise ValueError("Key must be specified to download file")

        f = self.S3.Object(bucket_name, key)
        f_delete = f.delete()

        return f_delete



'''pipeline.py - Tools for ruffus pipelines
===========================================

The :mod:`pipeline` module contains various utility functions for
interfacing CGAT ruffus pipelines with an HPC cluster, uploading data
to databases, providing parameterization, and more.

It is a collection of utility functions covering the topics:

* `pipeline control`_
* `Logging`_
* `Parameterisation`_
* `Running tasks`_
* `database upload`_
* `Report building`_

See :doc:`pipelines/pipeline_template` for a pipeline illustrating the
use of this module. See :ref:`pipelineSettingUp` on how to set up a
pipeline.

pipeline control
----------------

:mod:`pipeline` provides a :func:`main` function that provides command
line control to a pipeline. To use it, add::

    import cgatcore.pipeline as P
    # ...

    def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


    if __name__ == "__main__":
        sys.exit(P.main(sys.argv))

to your pipeline script. Typing::

    python my_pipeline.py --help

will provide the following output:

    Usage:
    usage: <pipeline> [OPTIONS] [CMD] [target]

    Execute pipeline mapping.

    Commands can be any of the following

    make <target>
        run all tasks required to build *target*

    show <target>
        show tasks required to build *target* without executing them

    plot <target>
        plot image (using inkscape) of pipeline state for *target*

    debug <target> [args]
        debug a method using the supplied arguments. The method <target>
        in the pipeline is run without checking any dependencies.

    config
        write new configuration files pipeline.yml with default values

    dump
        write pipeline configuration to stdout

    touch
        touch files only, do not run

    regenerate
        regenerate the ruffus checkpoint file

    check
        check if requirements (external tool dependencies) are satisfied.

    clone <source>
        create a clone of a pipeline in <source> in the current
        directory. The cloning process aims to use soft linking to files
        (not directories) as much as possible.  Time stamps are
        preserved. Cloning is useful if a pipeline needs to be re-run from
        a certain point but the original pipeline should be preserved.



    Options:
    --version             show program's version number and exit
    -h, --help            show this help message and exit
    --pipeline-action=PIPELINE_ACTION
                        action to take [default=none].
    --pipeline-format=PIPELINE_FORMAT
                        pipeline format [default=svg].
    -n, --dry-run         perform a dry run (do not execute any shell commands)
                        [default=False].
    -c CONFIG_FILE, --config-file=CONFIG_FILE
                        benchmark configuration file [default=pipeline.yml].
    -f FORCE_RUN, --force-run=FORCE_RUN
                        force running the pipeline even if there are up-to-
                        date tasks. If option is 'all', all tasks will be
                        rerun. Otherwise, only the tasks given as arguments
                        will be rerun. [default=False].
    -p MULTIPROCESS, --multiprocess=MULTIPROCESS
                        number of parallel processes to use on submit host
                        (different from number of jobs to use for cluster
                        jobs) [default=none].
    -e, --exceptions      echo exceptions immediately as they occur
                        [default=True].
    -i, --terminate       terminate immediately at the first exception
                        [default=none].
    -d, --debug           output debugging information on console, and not the
                        logfile [default=False].
    -s VARIABLES_TO_SET, --set=VARIABLES_TO_SET
                        explicitely set paramater values [default=[]].
    --input-glob=INPUT_GLOBS, --input-glob=INPUT_GLOBS
                        glob expression for input filenames. The exact format
                        is pipeline specific. If the pipeline expects only a
                        single input, `--input-glob=*.bam` will be sufficient.
                        If the pipeline expects multiple types of input, a
                        qualifier might need to be added, for example
                        `--input-glob=bam=*.bam` --input-glob=bed=*.bed.gz`.
                        Giving this option overrides the default of a pipeline
                        looking for input in the current directory or
                        specified the config file. [default=[]].
    --checksums=RUFFUS_CHECKSUMS_LEVEL
                        set the level of ruffus checksums[default=0].
    -t, --is-test         this is a test run[default=False].
    --engine=ENGINE       engine to use.[default=local].
    --always-mount        force mounting of arvados keep [False]
    --only-info           only update meta information, do not run
                        [default=False].
    --work-dir=WORK_DIR   working directory. Will be created if it does not
                        exist [default=none].
    --input-validation    perform input validation before starting
                        [default=False].

    pipeline logging configuration:
    --pipeline-logfile=PIPELINE_LOGFILE
                        primary logging destination.[default=pipeline.log].
    --shell-logfile=SHELL_LOGFILE
                        filename for shell debugging information. If it is not
                        an absolute path, the output will be written into the
                        current working directory. If unset, no logging will
                        be output. [default=none].

     Script timing options:
     --timeit=TIMEIT_FILE
                        store timeing information in file [none].
     --timeit-name=TIMEIT_NAME
                        name in timing file for this class of jobs [all].
     --timeit-header     add header for timing information [none].

     Common options:
     --random-seed=RANDOM_SEED
                        random seed to initialize number generator with
                        [none].
     -v LOGLEVEL, --verbose=LOGLEVEL
                        loglevel [1]. The higher, the more output.
     --log-config-filename=LOG_CONFIG_FILENAME
                        Configuration file for logger [logging.yml].
     --tracing=TRACING   enable function tracing [none].
     -?                  output short help (command line options only.

     cluster options:
      --no-cluster, --local
                        do no use cluster - run locally [False].
      --cluster-priority=CLUSTER_PRIORITY
                        set job priority on cluster [none].
      --cluster-queue=CLUSTER_QUEUE
                        set cluster queue [none].
      --cluster-num-jobs=CLUSTER_NUM_JOBS
                        number of jobs to submit to the queue execute in
                        parallel [none].
      --cluster-parallel=CLUSTER_PARALLEL_ENVIRONMENT
                        name of the parallel environment to use [none].
      --cluster-options=CLUSTER_OPTIONS
                        additional options for cluster jobs, passed on to
                        queuing system [none].
      --cluster-queue-manager=CLUSTER_QUEUE_MANAGER
                        cluster queuing system [sge].
      --cluster-memory-resource=CLUSTER_MEMORY_RESOURCE
                        resource name to allocate memory with [none].
      --cluster-memory-default=CLUSTER_MEMORY_DEFAULT
                        default amount of memory to allocate [unlimited].

    Input/output options:
     -I FILE, --stdin=FILE
                        file to read stdin from [default = stdin].
     -L FILE, --log=FILE
                        file with logging information [default = stdout].
     -E FILE, --error=FILE
                        file with error information [default = stderr].
     -S FILE, --stdout=FILE
                        file where output is to go [default = stdout].


Documentation on using pipelines is at :ref:`getting_started-Examples`.

Logging
-------

Logging is set up by :func:`main`. Logging messages will be sent to
the file :file:`pipeline.log` in the current directory.  Additionally,
messages are sent to an RabbitMQ_ message exchange to permit
monitoring of pipeline progress.

Running tasks
-------------

:mod:`pipeline` provides a :func:`pipeline.run` method to control
running commandline tools. The :func:`pipeline.run` method takes care
of distributing these tasks to the cluster. It takes into
consideration command line options such as ``--cluster-queue``. The
command line option ``--local`` will run jobs locally for testing
purposes.

For running Python code that is inside a module in a distributed
function, use the :func:`submit` function. The :func:`execute` method
runs a command locally.

Functions such as :func:`shellquote`, :func:`getCallerLocals`,
:func:`getCaller`, :func:`buildStatement`, :func:`expandStatement`,
:func:`joinStatements` support the parameter interpolation mechanism
used in :mod:`pipeline`.

Parameterisation
----------------

:mod:`pipeline` provides hooks for reading pipeline configuration
values from :file:`.ini` files and making them available inside ruffus_
tasks. The fundamental usage is a call to :func:`getParamaters` with
a list of configuration files, typically::

    # load options from the config file
    P.get_parameters(
        ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
        "../pipeline.yml",
        "pipeline.yml"])

The :mod:`pipeline` module defines a global variable :data:`PARAMS`
that provides access the configuration values. To get a handle to
this variable outside a pipeline script, call :func:`getParams`::

    my_cmd = "%(scripts_dir)s/bam2bam.py" % P.getParams()

Functions such as :func:`configToDictionary`, :func:`loadParameters`
:func:`matchParameter`, :func:`substituteParameters` support this
functionality.

Functions such as :func:`asList` and :func:`isTrue` are useful to work
with parameters.

The method :func:`peekParameters` allows one to programmatically read the
parameters of another pipeline.

Temporary files
---------------

Tasks containg multiple steps often require temporary memory storage
locations.  The functions :func:`getTempFilename`, :func:`getTempFile`
and :func:`getTempDir` provide these. These functions are aware of the
temporary storage locations either specified in configuration files or
on the command line and distinguish between the ``private`` locations
that are visible only within a particular compute node, and ``shared``
locations that are visible between compute nodes and typically on a
network mounted location.

Requirements
------------

The methods :func:`checkExecutables`, :func:`checkScripts` and
:func:`checkParameter` check for the presence of executables, scripts
or parameters. These methods are useful to perform pre-run checks
inside a pipeline if a particular requirement is met. But see also the
``check`` commandline command.

database upload
---------------

To assist with uploading data into a database, :mod:`pipeline` provides
several utility functions for conveniently uploading data. The :func:`load`
method uploads data in a tab-separated file::

    @transform("*.tsv.gz", suffix(".tsv.gz"), ".load")
    def loadData(infile, outfile):
        P.load(infile, outfile)

The methods :func:`mergeAndLoad` and :func:`concatenateAndLoad` upload
multiple files into same database by combining them first. The method
:func:`createView` creates a table or view derived from other tables
in the database. The function :func:`importFromIterator` uploads
data from a python list or other iterable directly.

The functions :func:`tablequote` and :func:`toTable` translate track
names derived from filenames into names that are suitable for tables.

The method :func:`build_load_statement` can be used to create an
upload command that can be added to command line statements to
directly upload data without storing an intermediate file.

The method :func:`connect` returns a database handle for querying the
database.

Package layout
--------------

The module is arranged as a python package with several submodules. Functions
within a submodule to be exported are all imported to the namespace of
:mod:`pipeline`.

.. toctree::

   cgatcore.pipeline.control
   cgatcore.pipeline.database
   cgatcore.pipeline.execution
   cgatcore.pipeline.files
   cgatcore.pipeline.parameters
   cgatcore.pipeline.utils


'''
import os

# import submodules into namespace
from cgatcore.pipeline.control import *
from cgatcore.pipeline.database import *
from cgatcore.pipeline.files import *
from cgatcore.pipeline.cluster import *
from cgatcore.pipeline.execution import *
from cgatcore.pipeline.utils import *
from cgatcore.pipeline.parameters import *


# __all__ = [
#     # backwards incompatibility
#     "clone",
#     "touch",
#     "snip",
#     # execution.py
#     "run",
#     "execute",
#     "shellquote",
#     "buildStatement",
#     "submit",
#     "joinStatements",
#     "cluster_runnable",
#     "run_pickled",
#     # database.py
#     "tablequote",
#     "toTable",
#     "build_load_statement",
#     "load",
#     "concatenateAndLoad",
#     "mergeAndLoad",
#     "connect",
#     "createView",
#     "getdatabaseName",
#     "importFromIterator",
#     # Utils.py
#     "add_doc",
#     "isTest",
#     "getCallerLocals",
#     "getCaller",
#     # Control.py
#     "main",
#     "peekParameters",
#     # Files.py
#     "getTempFile",
#     "getTempDir",
#     "getTempFilename",
#     "checkScripts",
#     "checkExecutables",
#     # Local.py
#     "run_report",
#     "publish_report",
#     "publish_notebooks",
#     "publish_tracks",
#     "getProjectDirectories",
#     "getpipelineName",
#     "getProjectId",
#     "getProjectName",
#     "isCGAT",
#     # Parameters.py
#     "getParameters",
#     "loadParameters",
#     "matchParameter",
#     "substituteParameters",
#     "asList",
#     "checkParameter",
#     "isTrue",
#     "configToDictionary",
# ]



import os
import sys
from abc import abstractmethod


class AbstractRemoteObject():
    '''This is an abstract class that all RemoteObjects will
       inherit from. This is an abstract class to ridgidly define
       the abstract methods of this RemoteObject class'''

    def __init__(self, *args, **kwargs):

        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def exists(self):
        pass

    @abstractmethod
    def download(self):
        pass

    @abstractmethod
    def upload(self):
        pass

    @abstractmethod
    def delete_file(self):
        pass



import os
import sys

try:
    import azure
except Exception:
    raise ImportError(
        "The azure-batch and azure-mgmt-storage packages needs to be installed.")

from azure.storage.blob import BlockBlobService, PublicAccess
from cgatcore.remote import AbstractRemoteObject


class AzureRemoteObject(AbstractRemoteObject):
    '''This is a class that will interact with an AWS object store.'''

    def __init__(self, account_name, account_key, *args, **kwargs):
        super(AzureRemoteObject, self).__init__(*args, **kwargs)

        self.account_name = account_name
        self.account_key = account_key

        self._Azureobject = AzureConnection(
            self.account_name, self.account_key, *args, **kwargs)

    def exists(self, bucket_name):

        try:
            self._Azureobject.bucket_exists(bucket_name)
        except Exception:
            raise KeyError("No bucket exists, please specify a valid bucket")
        return True

    def download(self, container_name, blob_name, file_dir):
        self._Azureobject.remote_download(container_name, blob_name, file_dir)
        os.sync()  # ensure flush to disk
        return file_dir

    def upload(self, container_name, blob_name, file_dir):
        self._Azureobject.remote_upload(container_name, file_dir, blob_name)
        return file_dir

    def delete_file(self, container_name, blob_name):
        self._Azureobject.remote_delete_file(container_name, blob_name)
        return blob_name


class AzureConnection():
    '''This is a connection to a remote Azure bucket for Microsoft Azure
    platform using the azure SDK.'''

    def __init__(self, account_name, account_key, *args, **kwargs):

        self.Azure = BlockBlobService(
            account_name=account_name, account_key=account_key, **kwargs)

    def bucket_exists(self, container_name):
        'In microsoft buckets are called containers'

        # To Do: rename function because of discrep[ancy between bucket and
        # containers - may need to alter the name to object storage?
        try:
            self.Azure.exists(container_name)
            return True
        except Exception:
            return False

    def remote_download(self,
                        container_name,
                        blob_name,
                        dest_dir):
        '''Download data/file from an Azure container.'''

        if not container_name:
            raise ValueError(
                "Container name must be specified to download file")
        if not blob_name:
            raise ValueError("Blob name must be specified to download file")

        if dest_dir:
            dest_path = os.path.realpath(os.path.expanduser(dest_dir))
        try:
            self.Azure.get_blob_to_path(container_name, blob_name, dest_path)
        except Exception:
            raise Exception('''No file was downloaded, make sure the correct
                            file or path is specified. It currently is: {}'''.format(dest_path))

    def remote_upload(self,
                      container_name,
                      file_dir,
                      blob_name=None):
        '''Upload data/file to an Azure container.'''

        file_path = os.path.realpath(os.path.expanduser(file_dir))

        if not container_name:
            raise ValueError("Container name must be specified to upload file")
        if not os.path.exists(file_dir):
            raise ValueError(
                "File path specified does not exitis: {}".format(file_path))
        if not os.path.isfile(file_dir):
            raise ValueError(
                "File path specified is not a file: {}".format(file_path))

        if not self.bucket_exists(container_name):
            self.Azure.create_container(container_name)

        try:
            self.Azure.create_blob_from_path(
                container_name, blob_name, file_dir)
        except Exception:
            raise Exception(
                "filename is not correctly specified: {}".format(file_dir))

    def remote_delete_file(self, container_name, blob_name):
        '''Will remove the blob from the remote Azure container'''

        if not container_name:
            raise ValueError("Bucket name must be specified to download file")
        if not blob_name:
            raise ValueError("Key must be specified to download file")

        try:
            self.Azure.delete_blob(container_name, blob_name)
        except Exception:
            raise Exception(
                "No deletion of blob took place, please make sure the container_name and/or blob_name are correct")
